<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0">

<!--
  This file contains a bunch of generic routines that are used by the
  other xsl files.  This file should only ever be included by other
  xsl files and not run directly.
  -->

  <xsl:output method="text"/>


  <!--
    Define some global variables.
    -->
    
  <!-- Makes the code cleaner if you have newline variables -->
  <xsl:variable name="newline">
<xsl:text>
</xsl:text>
  </xsl:variable>
  
  <xsl:variable name="dblnewline">
<xsl:text>

</xsl:text>
  </xsl:variable>

  <!-- The name of the user preferences xml file -->
  <xsl:variable name="userxmlfilename">
    <xsl:text>user.xml</xsl:text>
  </xsl:variable>

  <!-- These are all the types that we want void* arguments to be expanded
    to. "i" means integer, "r" means real -->
  <xsl:variable name="datatypes" select="'ir'"/>



  <!--
    This checks the length of an indentifier and displays a warning
    message if the length is greater than the max length specified in 
    the user options xml file.
    -->
  <xsl:template name="check_identifier_length">
    <xsl:param name="identifier"/>

    <xsl:variable name="max_len" select="document($userxmlfilename)/user/options/option[@name = 'max_identifier_length']/@length"/>

    <xsl:if test="string-length($identifier) &gt; $max_len">
      <xsl:message>
        WARNING: identifier length is greater than <xsl:value-of select="$max_len"/> characters: <xsl:value-of select="$identifier"/>
      </xsl:message>
    </xsl:if>

  </xsl:template>



  <!--
    Generate a type name for a fortran type.  The tricky part of this task
    is to figure out if a type needs to have its class name prepended to
    the type name, and how to find the proper class name to prepend.
    
    All types get their class name (if available) prepended to the type
    name in order to avoid cross-class type name clashes.
    -->
  <xsl:template name="gen_type_name_f">
    <xsl:param name="append_type" select="'no'"/>
    <xsl:param name="type_name"/>

    <!-- since we need to check the resulting length of the generated name,
         we do all the output of this template into a variable, which we can
         just then check at the end -->
    <xsl:variable name="final_name">

      <!-- if it already contains a class name, then we use that one -->
      <xsl:if test="contains( $type_name, '::' )">
        <xsl:call-template name="parse_type_name_f">
          <xsl:with-param name="type_name" select="$type_name"/>
        </xsl:call-template>
      </xsl:if>
            
      <!-- if it does not already contain a class name, then we have a
           little more work to do to figure out what the name is -->
      <xsl:if test="not(contains( $type_name, '::' ))">
        <xsl:call-template name="lookup_type_name_f">
          <xsl:with-param name="type_name" select="$type_name"/>
        </xsl:call-template>
      </xsl:if>
   
      <!-- check the parameter to see if we should append "_type" -->
      <xsl:if test="$append_type = 'yes'">
        <xsl:text>_type</xsl:text>
      </xsl:if>

    </xsl:variable>

    <!-- use the dictionary to rename/shorten the final type name -->
<!--    <xsl:variable name="fixed_name">
      <xsl:call-template name="fixname">
        <xsl:with-param name="name" select="$final_name"/>
      </xsl:call-template>
    </xsl:variable>
-->
    <!-- so now we created our variable, so output it -->
    <xsl:value-of select="$final_name"/>

    <!-- and check to make sure that the length is legal -->
    <xsl:call-template name="check_identifier_length">
      <xsl:with-param name="identifier" select="$final_name"/>
    </xsl:call-template>
    
  </xsl:template>



  <!--
    This template takes a type name of the form classname::typename and
    converts it into a legal name for fortran.  It also applies any name
    shortening (from the user preferences file) that is appropriate.
    -->
  <xsl:template name="parse_type_name_f">
    <xsl:param name="type_name"/>

    <!-- find out what the original class name is -->
    <xsl:variable name="class_name" select="substring-before( $type_name, '::' )"/>

    <!-- find out what the original typedef name is -->
    <xsl:variable name="type_orig" select="substring-after( $type_name, '::' )"/>

    <!-- lookup the aliases in the user xml file -->
    <xsl:variable name="class_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@alias"/>
    <xsl:variable name="type_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/typedef[@name = $type_orig]/@alias"/>

    <!-- if we found a class alias, then use that -->
    <xsl:value-of select="$class_alias"/>
    <xsl:if test="string-length( $class_alias ) = 0">
      <xsl:value-of select="$class_name"/>
    </xsl:if>

    <xsl:text>_</xsl:text>

    <!-- if we fould a type alias, then use that -->
    <xsl:value-of select="$type_alias"/>
    <xsl:if test="string-length( $type_alias ) = 0">
      <xsl:value-of select="$type_orig"/>
    </xsl:if>

  </xsl:template>



  <!--
    Given a type name, this template locates where the type is defined,
    and then generates its full name accordingly.  It starts searching
    the current class for a definition of the type, and if it doesn't
    find it, it continues searching up any nested structures looking for
    it.

    The real problem is that it's hard to tell if a type is defined in
    a class and needs to be qualified, or if it is some other externally
    defined type.
    
    Ideally, the way this should work is that it should first check the
    current structure context to see if the type is defined within the
    current class.  If it is, then we know that we can just prepend the
    current class name (with any renaming/shortening of course).  If it
    isn't, then we need to check the next upmost structure to see if the
    current structure is defined within another, and if that parent has
    the type defined within it.  If we get through all the nested
    structures and still haven't found the type, then we should check
    the global structure context to see if it is there (just to make
    sure I guess, but this last step is not necessarily necessary).
    -->
  <xsl:template name="lookup_type_name_f">
    <xsl:param name="type_name"/>

    <xsl:choose>
    
    <!-- if the type is defined within the current structure context, then
         just prepend the class name to the type name. --> 
    <xsl:when test="parent::structure/typealias[@name = $type_name] or parent::structure/structure[@name = $type_name]">
      
      <!-- get the original class name into the class_name variable -->
      <xsl:variable name="class_name" select="parent::structure/@name"/>
    
      <!-- lookup the aliases in the user preferences file -->
      <xsl:variable name="class_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/@alias"/>
      <xsl:variable name="type_alias" select="document($userxmlfilename)/user/classes/class[@name = $class_name]/typedef[@name = $type_name]/@alias"/>

      <!-- if the class name is the same as the type name, then we
           don't prepend the classname -->
      <xsl:if test="$class_name != $type_name">

        <!-- if we found a class alias, then use that -->
        <xsl:value-of select="$class_alias"/>
        <xsl:if test="string-length( $class_alias ) = 0">
          <xsl:value-of select="$class_name"/>
        </xsl:if>

        <xsl:text>_</xsl:text>

      </xsl:if>

      <!-- if we fould a type alias, then use that -->
      <xsl:value-of select="$type_alias"/>
      <xsl:if test="string-length( $type_alias ) = 0">
        <xsl:value-of select="$type_name"/>
      </xsl:if>

    </xsl:when>

   <!-- if the type is not defined within the current structure context, then
        figure out where the type is defined and prepend any class names as
        necessary.

        but, since i'm leaving in a week, we'll just assume that if the
        type is not defined within the immediate class, then it's global.
        and needs no class name prepended.  -->
    <xsl:otherwise>
      <xsl:variable name="type_alias" select="document($userxmlfilename)/user/classes/class[@name = $type_name]/@alias"/>

      <xsl:value-of select="$type_alias"/>
      <xsl:if test="string-length( $type_alias ) = 0">
        <xsl:value-of select="$type_name"/>
      </xsl:if>
    </xsl:otherwise>

    </xsl:choose>
  </xsl:template>



  <!--
    This template renames names (methods, types, arguments, etc).  This is 
    accomplished through the use of a dictionary of word/abbreviation pairs
    stored in the user xml file.  The first occurence of each word in the 
    dictionary that appears in the passed in string is replaced with its
    abbreviation.
    
    Since xsl doesn't have loops and has limited string manipulation
    routines, the string substitution must be accomplished via recursion.
    As a result, this template runs extremely slow.  Ideally the xsl
    shouldn't do this and have it done in some other more appropriate way.
    
    Also, this template only replaces a single occurence of each dictionary
    term, so if you want to replace more than one occurence of a term, add
    it to the dictionary twice (or however many times).
    -->
  <xsl:template name="fixname_orig">
    <xsl:param name="name"/>
   
      <xsl:apply-templates select="document($userxmlfilename)/user/substitutions/sub" mode="fixname_helper">
        <xsl:with-param name="name" select="$name"/>
        <xsl:with-param name="pos" select="1"/>
      </xsl:apply-templates>
   
  </xsl:template>

  <xsl:template name="fixname">
    <xsl:param name="name"/>
   

   <xsl:value-of select="$name"/>
   
  </xsl:template>



  <!--
    This is the helper template for the fixname() template.
    -->
  <xsl:template match="sub" mode="fixname_helper">
    <xsl:param name="name"/>
    <xsl:param name="pos"/>
      
    <!-- this template matches all the elements in the dictionary, but we
         just want to step through them one by one, so we check if this
         dictionary element is the one we're after. -->
    <xsl:if test="position() = $pos">  

      <!-- create a new variable that will store the new abbreviated name -->
      <xsl:variable name="newname">

      <!-- if the name contains the dictionary term -->
      <xsl:choose>
      <xsl:when test="contains( $name, @original )">
        <!-- construct the new name -->
        <xsl:value-of select="substring-before( $name, @original )"/>
        <xsl:value-of select="@new"/>
        <xsl:value-of select="substring-after( $name, @original )"/>
      </xsl:when>
      
      <!-- if the name does not contain the dictionary term -->
      <xsl:otherwise>
        <xsl:value-of select="$name"/>  
      </xsl:otherwise>
      </xsl:choose>
    
      </xsl:variable>
      
      <!-- at this point, we have (a possibly abbreviated form of) the name
           in a variable called newname -->
      
      <xsl:choose>
      <!-- if this is the last term in the dictionary, stop the recursion -->
      <xsl:when test="$pos = count( document($userxmlfilename)/user/substitutions/sub )">
        <xsl:value-of select="$newname"/>
      </xsl:when>

      <!-- if this is not the last term in the dictionary, then make the
           recursive call to check the next term in the dictionary -->
      <xsl:otherwise>
        <xsl:apply-templates select="document($userxmlfilename)/user/substitutions/sub" mode="fixname_helper">
          <xsl:with-param name="pos" select="$pos + 1"/>
          <xsl:with-param name="name" select="$newname"/>
        </xsl:apply-templates>
      </xsl:otherwise>
      </xsl:choose>
          
    </xsl:if>
  </xsl:template>



  <!--
    This checks if a method is invalid in some way and shouldn't be
    wrapped.  Right now it considers these methods invalid:
      o operator overloading methods
      o private methods
      o methods with std::ostream arguments
      o methods with std::istream arguments
      o methods with function pointer arguments
      o methods with specified 2d array bounds
      o methods called "Scale", "IsTypeOf", "Transpose"
      
    Since we want to pass a result back to the caller, the way we do
    that is to simply output 'pass' or 'fail' and the caller should
    wrap the call to this template in a variable, then the caller can
    just check the value of the variable to see if the method passed
    the check or not.
    -->
  <xsl:template name="check_method">

    <xsl:choose>
      
      <!-- if this method is an overloaded operator, then fail -->
      <xsl:when test="@op = 'yes'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
  
      <!-- if this is not a public method, then fail -->
      <xsl:when test="@access != 'pub'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
  
      <!-- if this method has a std::ostream argument, then fail -->
      <xsl:when test="descendant::type/@usertype='std::ostream'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method has a std::istream argument, then fail -->
      <xsl:when test="descendant::type/@usertype='std::istream'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method has a function pointer, then fail -->
      <xsl:when test="descendant::type/@kind='func'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method is named "Scale", then fail.  this really should
           convert both comparators to uppercase, since fortran is not
           case sensitive. -->
      <xsl:when test="@name = 'Scale'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method is named "IsTypeOf", then fail. this is to get
           around a bug in which too many copies (due to inheritance)
           of this wrapper are being generated. -->
      <xsl:when test="@name = 'IsTypeOf'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method is named "Transpose", then fail. -->
      <xsl:when test="@name = 'Transpose'">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if this method has a specified 2d bound, e.g. int x[3][3], then
           fail. note that these arguments are detected by simply looking
           for an array element that has a rank attribute.  this assumes
           that only args with specified bounds have rank attributes, but
           this may not be the case... -->
      <xsl:when test="descendant::array/@rank">
        <xsl:text>fail</xsl:text>
      </xsl:when>
      
      <!-- if none of the above are true, then the method passes -->
      <xsl:otherwise>
        <xsl:text>pass</xsl:text>
      </xsl:otherwise>
  
    </xsl:choose>

  </xsl:template>



  <!--
    This is a recursive template that returns just the filename part of a
    path.  e.g: /Users/tomb/Projects/readme.txt becomes readme.txt
   -->
  <xsl:template name="getfilename">
    <xsl:param name="path"/>
  
    <xsl:choose>
      
      <!-- if there is a / in the string, then make a recursive call on
           the part of the string after the /. -->
      <xsl:when test="contains( $path, '/' )">
        <xsl:call-template name="getfilename">
          <xsl:with-param name="path" select="substring-after( $path, '/' )"/>
        </xsl:call-template>
      </xsl:when>
      
      <!-- if there is no / in the string, then we're finished. -->
      <xsl:otherwise>
        <xsl:value-of select="$path"/>
      </xsl:otherwise>
      
    </xsl:choose>
  </xsl:template>



</xsl:stylesheet>
